package java.awt;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;
import sun.util.logging.PlatformLogger;
import java.io.Serializable;

public class ContainerOrderFocusTraversalPolicy extends FocusTraversalPolicy implements Serializable
{
    private static final PlatformLogger log;
    private final int FORWARD_TRAVERSAL = 0;
    private final int BACKWARD_TRAVERSAL = 1;
    private static final long serialVersionUID = 486933713763926351L;
    private boolean implicitDownCycleTraversal;
    private transient Container cachedRoot;
    private transient List<Component> cachedCycle;
    
    public ContainerOrderFocusTraversalPolicy() {
        this.implicitDownCycleTraversal = true;
    }
    
    private List<Component> getFocusTraversalCycle(Container container) {
        ArrayList<Component> list = new ArrayList();
        this.enumerateCycle(container, list);
        return list;
    }
    
    private int getComponentIndex(List<Component> list, Component component) {
        return list.indexOf(component);
    }
    
    private void enumerateCycle(Container container, List<Component> list) {
        if (!container.isVisible() || !container.isDisplayable()) {
            return;
        }
        list.add(container);
        Component[] components = container.getComponents();
        for (int i = 0; i < components.length; ++i) {
            Component component = components[i];
            if ((component instanceof Container)) {
                Container container2 = (Container)component;
                if (!container2.isFocusCycleRoot() && !container2.isFocusTraversalPolicyProvider()) {
                    this.enumerateCycle(container2, list);
                    continue;
                }
            }
            list.add(component);
        }
        return;
    }
    
    private Container getTopmostProvider(Container container, Component component) {
        Container container2 = component.getParent();
        Container container3 = null;
        while (container2 != container && container2 != null) {
            if (container2.isFocusTraversalPolicyProvider()) {
                container3 = container2;
            }
            container2 = container2.getParent();
        }
        if (container2 == null) {
            return null;
        }
        return container3;
    }
    
    private Component getComponentDownCycle(Component component, int n) {
        Component defaultComponent = null;
        if ((component instanceof Container)) {
            Container container = (Container)component;
            if (container.isFocusCycleRoot()) {
                if (this.getImplicitDownCycleTraversal()) {
                    defaultComponent = container.getFocusTraversalPolicy().getDefaultComponent(container);
                    if (defaultComponent != null && ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                        ContainerOrderFocusTraversalPolicy.log.fine("### Transfered focus down-cycle to " + defaultComponent + " in the focus cycle root " + container);
                    }
                }
                else {
                    return null;
                }
            }
            else if (container.isFocusTraversalPolicyProvider()) {
                defaultComponent = ((n == 0) ? container.getFocusTraversalPolicy().getDefaultComponent(container) : container.getFocusTraversalPolicy().getLastComponent(container));
                if (defaultComponent != null && ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Transfered focus to " + defaultComponent + " in the FTP provider " + container);
                }
            }
        }
        return defaultComponent;
    }
    
    @Override
    public Component getComponentAfter(Container obj, Component component) {
        if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
            ContainerOrderFocusTraversalPolicy.log.fine("### Searching in " + obj + " for component after " + component);
        }
        if (obj == null || component == null) {
            throw new IllegalArgumentException("aContainer and aComponent cannot be null");
        }
        if (!obj.isFocusTraversalPolicyProvider() && !obj.isFocusCycleRoot()) {
            throw new IllegalArgumentException("aContainer should be focus cycle root or focus traversal policy provider");
        }
        if (obj.isFocusCycleRoot() && !component.isFocusCycleRoot(obj)) {
            throw new IllegalArgumentException("aContainer is not a focus cycle root of aComponent");
        }
        synchronized (obj.getTreeLock()) {
            if (!obj.isVisible() || !obj.isDisplayable()) {
                return null;
            }
            Component componentDownCycle = this.getComponentDownCycle(component, 0);
            if (componentDownCycle != null) {
                return componentDownCycle;
            }
            Container topmostProvider = this.getTopmostProvider(obj, component);
            if (topmostProvider != null) {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Asking FTP " + topmostProvider + " for component after " + component);
                }
                Component componentAfter = topmostProvider.getFocusTraversalPolicy().getComponentAfter(topmostProvider, component);
                if (componentAfter != null) {
                    if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                        ContainerOrderFocusTraversalPolicy.log.fine("### FTP returned " + componentAfter);
                    }
                    return componentAfter;
                }
                else {
                    component = topmostProvider;
                }
            }
            List<Component> focusTraversalCycle = this.getFocusTraversalCycle(obj);
            if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                ContainerOrderFocusTraversalPolicy.log.fine("### Cycle is " + focusTraversalCycle + ", component is " + component);
            }
            int i = this.getComponentIndex(focusTraversalCycle, component);
            if (i < 0) {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Didn't find component " + component + " in a cycle " + obj);
                }
                return this.getFirstComponent(obj);
            }
            else {
                ++i;
                while (i < focusTraversalCycle.size()) {
                    Component component3 = (Component)focusTraversalCycle.get(i);
                    if (this.accept(component3)) {
                        return component3;
                    }
                    Component componentDownCycle2;
                    if ((componentDownCycle2 = this.getComponentDownCycle(component3, 0)) != null) {
                        return componentDownCycle2;
                    }
                    ++i;
                }
                if (obj.isFocusCycleRoot()) {
                    this.cachedRoot = obj;
                    this.cachedCycle = focusTraversalCycle;
                    Component firstComponent = this.getFirstComponent(obj);
                    this.cachedRoot = null;
                    this.cachedCycle = null;
                    return firstComponent;
                }
            }
        }
        return null;
    }
    
    @Override
    public Component getComponentBefore(Container container, Component obj) {
        if (container == null || obj == null) {
            throw new IllegalArgumentException("aContainer and aComponent cannot be null");
        }
        if (!container.isFocusTraversalPolicyProvider() && !container.isFocusCycleRoot()) {
            throw new IllegalArgumentException("aContainer should be focus cycle root or focus traversal policy provider");
        }
        if (container.isFocusCycleRoot() && !obj.isFocusCycleRoot(container)) {
            throw new IllegalArgumentException("aContainer is not a focus cycle root of aComponent");
        }
        synchronized (container.getTreeLock()) {
            if (!container.isVisible() || !container.isDisplayable()) {
                return null;
            }
            Container topmostProvider = this.getTopmostProvider(container, obj);
            if (topmostProvider != null) {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Asking FTP " + topmostProvider + " for component after " + obj);
                }
                Component componentBefore = topmostProvider.getFocusTraversalPolicy().getComponentBefore(topmostProvider, obj);
                if (componentBefore != null) {
                    if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                        ContainerOrderFocusTraversalPolicy.log.fine("### FTP returned " + componentBefore);
                    }
                    return componentBefore;
                }
                else {
                    obj = topmostProvider;
                    if (this.accept(obj)) {
                        return obj;
                    }
                }
            }
            List<Component> focusTraversalCycle = this.getFocusTraversalCycle(container);
            if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                ContainerOrderFocusTraversalPolicy.log.fine("### Cycle is " + focusTraversalCycle + ", component is " + obj);
            }
            int i = this.getComponentIndex(focusTraversalCycle, obj);
            if (i < 0) {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Didn't find component " + obj + " in a cycle " + container);
                }
                return this.getLastComponent(container);
            }
            else {
                --i;
                while (i >= 0) {
                    Component component3 = (Component)focusTraversalCycle.get(i);
                    Component componentDownCycle;
                    if (component3 != container && (componentDownCycle = this.getComponentDownCycle(component3, 1)) != null) {
                        return componentDownCycle;
                    }
                    if (this.accept(component3)) {
                        return component3;
                    }
                    --i;
                }
                if (container.isFocusCycleRoot()) {
                    this.cachedRoot = container;
                    this.cachedCycle = focusTraversalCycle;
                    Component lastComponent = this.getLastComponent(container);
                    this.cachedRoot = null;
                    this.cachedCycle = null;
                    return lastComponent;
                }
            }
        }
        return null;
    }
    
    @Override
    public Component getFirstComponent(Container obj) {
        if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
            ContainerOrderFocusTraversalPolicy.log.fine("### Getting first component in " + obj);
        }
        if (obj == null) {
            throw new IllegalArgumentException("aContainer cannot be null");
        }
        synchronized (obj.getTreeLock()) {
            if (!obj.isVisible() || !obj.isDisplayable()) {
                return null;
            }
            List<Component> obj2;
            if (this.cachedRoot == obj) {
                obj2 = this.cachedCycle;
            }
            else {
                obj2 = this.getFocusTraversalCycle(obj);
            }
            if (obj2.size() == 0) {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Cycle is empty");
                }
                return null;
            }
            else {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Cycle is " + obj2);
                }
                for (Component component : obj2) {
                    if (this.accept(component)) {
                        return component;
                    }
                    Component componentDownCycle;
                    if (component != obj && (componentDownCycle = this.getComponentDownCycle(component, 0)) != null) {
                        return componentDownCycle;
                    }
                    continue;
                }
            }
        }
        return null;
    }
    
    @Override
    public Component getLastComponent(Container obj) {
        if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
            ContainerOrderFocusTraversalPolicy.log.fine("### Getting last component in " + obj);
        }
        if (obj == null) {
            throw new IllegalArgumentException("aContainer cannot be null");
        }
        synchronized (obj.getTreeLock()) {
            if (!obj.isVisible() || !obj.isDisplayable()) {
                return null;
            }
            List<Component> obj2;
            if (this.cachedRoot == obj) {
                obj2 = this.cachedCycle;
            }
            else {
                obj2 = this.getFocusTraversalCycle(obj);
            }
            if (obj2.size() == 0) {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Cycle is empty");
                }
                return null;
            }
            else {
                if (ContainerOrderFocusTraversalPolicy.log.isLoggable(PlatformLogger.Level.FINE)) {
                    ContainerOrderFocusTraversalPolicy.log.fine("### Cycle is " + obj2);
                }
                int i = obj2.size() - 1;
                while (i >= 0) {
                    Component component = (Component)obj2.get(i);
                    if (this.accept(component)) {
                        return component;
                    }
                    if (component instanceof Container && component != obj) {
                        Container container = (Container)component;
                        if (container.isFocusTraversalPolicyProvider()) {
                            Component lastComponent = container.getFocusTraversalPolicy().getLastComponent(container);
                            if (lastComponent != null) {
                                return lastComponent;
                            }
                        }
                    }
                    --i;
                }
            }
        }
        return null;
    }
    
    @Override
    public Component getDefaultComponent(Container container) {
        return this.getFirstComponent(container);
    }
    
    public void setImplicitDownCycleTraversal(boolean implicitDownCycleTraversal) {
        this.implicitDownCycleTraversal = implicitDownCycleTraversal;
    }
    
    public boolean getImplicitDownCycleTraversal() {
        return this.implicitDownCycleTraversal;
    }
    
    protected boolean accept(Component component) {
        if (!component.canBeFocusOwner()) {
            return false;
        }
        if (!(component instanceof Window)) {
            Container container = component.getParent();
            while (container != null) {
                if (!container.isEnabled() && !container.isLightweight()) {
                    return false;
                }
                if ((container instanceof Window)) {
                    break;
                }
                container = container.getParent();
            }
        }
        return true;
    }
    
    static {
        log = PlatformLogger.getLogger("java.awt.ContainerOrderFocusTraversalPolicy");
    }
}
